
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <algorithm>
#include <vector>
#include <iostream>
#include <tclap/CmdLine.h>

#include "util.h"
#include "const.h"
#include "render-region.h"
#include "call_pixie.h"
#include "config.h"

using namespace std;

int GROUND_TRUTH_SUPER_SAMPLING = 3;
int GROUND_TRUTH_TIME_SAMPLING_RATE = 6;

// Command line arguments
TCLAP::CmdLine *cmd = NULL;
TCLAP::CmdLine &CmdLineSingleton() {
  if (!cmd) cmd = new TCLAP::CmdLine("PRT Shader", ' ', "0.1");
  return *cmd;
}

TCLAP::ValueArg<int> spatial_sampling_rate_arg(
    "s","spatial_sampling_rate", "Spatial Sampling Rate", false, 4, "int", CmdLineSingleton());

TCLAP::ValueArg<int> temporal_sampling_rate_arg(
    "t","temporal_sampling_rate", "Temporal Sampling Rate", false, 4, "int", CmdLineSingleton());

TCLAP::ValueArg<string> config_filename_arg(
    "c","config", "Config filename", false, "quad.config", "string", CmdLineSingleton());

TCLAP::SwitchArg use_sinc_filter_arg(
    "","sinc","Whether to use Sinc function as the filter.", CmdLineSingleton(), false);

TCLAP::SwitchArg keep_sample_files_arg(
    "","keepfiles","Whether to keep temp sample files.", CmdLineSingleton(), false);

TCLAP::ValueArg<string> output_filename_arg(
    "o","output", "Output filename", false, "", "string", CmdLineSingleton());

TCLAP::ValueArg<string> manual_render_region_arg(
    "","mrr", "Manual Render Region", false, "", "string", CmdLineSingleton());

TCLAP::ValueArg<double> animation_start_arg(
    "","animationstart", "Start (relative to .rib frames) of animation", false, -1, "double", CmdLineSingleton());

TCLAP::ValueArg<double> animation_length_arg(
    "","animationlength", "Frames of animation for this image (relative to .rib frames)", false, -1, "double", CmdLineSingleton());
TCLAP::ValueArg<int> scanlines_arg(
    "","scanlines", "Scanlines per call to pixie", false, 50, "int", CmdLineSingleton());


vector<InputSample> inSamples;
vector<ColorSample> outSamples;

int main(int argc, char* argv[]) {
  CmdLineSingleton().parse( argc, argv );
  LoadConfig(config_filename_arg.getValue().c_str());
  FilterFunc filter_func =
      use_sinc_filter_arg.getValue() ?
      SincWindow : GaussianWindow;

  if (manual_render_region_arg.getValue().length() == 0) {
    RENDER_TOP = 0;
    RENDER_BOTTOM = IMAGE_HEIGHT - 1;
  } else {
    setRenderRegionString(manual_render_region_arg.getValue().c_str());
  }

  GROUND_TRUTH_TIME_SAMPLING_RATE = temporal_sampling_rate_arg.getValue();
  GROUND_TRUTH_SUPER_SAMPLING = spatial_sampling_rate_arg.getValue();

  float* img = new float[IMAGE_WIDTH * IMAGE_HEIGHT * 4];
  float* img_tmp_out = new float[IMAGE_WIDTH * IMAGE_HEIGHT * 4];
  ClearImage(img, IMAGE_WIDTH, IMAGE_HEIGHT);
  ClearImage(img_tmp_out, IMAGE_WIDTH, IMAGE_HEIGHT);

  float weight_sum[IMAGE_WIDTH][IMAGE_HEIGHT];
  for (int x = 0; x < IMAGE_WIDTH; x++) {
    for (int y = 0; y < IMAGE_HEIGHT; y++) {
      weight_sum[x][y] = 0;
    }
  }

  char out_filename[1024];
  if (output_filename_arg.getValue().length() == 0) {
    sprintf(out_filename, "brute_force_%dx%dx%d.exr",
            spatial_sampling_rate_arg.getValue(),
            spatial_sampling_rate_arg.getValue(),
            temporal_sampling_rate_arg.getValue());
  } else {
    strcpy(out_filename, output_filename_arg.getValue().c_str());
  }

  const int batch_scanlines = scanlines_arg.getValue();
  for (int render_y = RENDER_TOP; render_y <= RENDER_BOTTOM; render_y += batch_scanlines) {
    inSamples.clear();
    outSamples.clear();
    outSamples.reserve((int)
                   (IMAGE_WIDTH * GROUND_TRUTH_SUPER_SAMPLING *
                    GROUND_TRUTH_SUPER_SAMPLING *
                    GROUND_TRUTH_TIME_SAMPLING_RATE *
                    (TIME_END - TIME_START)));


    int last_scanline = MIN(render_y + batch_scanlines, RENDER_BOTTOM);

    // Generate 1st pass samples.
    GenerateSamples(IMAGE_WIDTH, IMAGE_HEIGHT,
                    TIME_START, TIME_END,
                    IMAGE_WIDTH * GROUND_TRUTH_SUPER_SAMPLING,
                    IMAGE_HEIGHT * GROUND_TRUTH_SUPER_SAMPLING,
                    (int)(GROUND_TRUTH_TIME_SAMPLING_RATE * (TIME_END - TIME_START)),
                    0, IMAGE_WIDTH,
                    render_y, last_scanline,   // Render region: a scan line
                    true, inSamples, true);

    cout << "render y = " << render_y << ", samples " << inSamples.size() << endl;

    char input_filename[1024];
    char output_filename[1024];
    sprintf(input_filename, "/tmp/br_inputfile-%d-%d-%d",
            RENDER_TOP, RENDER_BOTTOM, render_y);
    sprintf(output_filename, "/tmp/br_outputfile-%d-%d-%d",
            RENDER_TOP, RENDER_BOTTOM, render_y);

    if (animation_start_arg.getValue() > 0 && animation_length_arg.getValue() > 0) {
        RenderColorSamplesMultiFile(inSamples, outSamples, IMAGE_WIDTH, IMAGE_HEIGHT,
            RIB_FILENAME, input_filename, output_filename, true,
            keep_sample_files_arg.getValue(),
            animation_start_arg.getValue(), 1.0 / animation_length_arg.getValue());
    } else {
        // Render 1st pass samples.
        RenderColorSamples(inSamples, outSamples, IMAGE_WIDTH, IMAGE_HEIGHT,
            RIB_FILENAME,
            input_filename, output_filename, true, keep_sample_files_arg.getValue());
    }

    const int SEARCH_RADIUS = 4;
    for (int i = 0; i < outSamples.size(); i++) {
      ColorSample &sample = outSamples[i];
      for (int x = (int) (sample.x - SEARCH_RADIUS + 0.5);
               x <= (int) (sample.x + SEARCH_RADIUS - 0.50001); x++) {
        for (int y = (int) (sample.y - SEARCH_RADIUS + 0.5);
                 y <= (int) (sample.y + SEARCH_RADIUS - 0.50001); y++) {
          if (x < 0 || y < 0 || x >= IMAGE_WIDTH || y >= IMAGE_HEIGHT)
            continue;
          float center_x = x + 0.5;
          float center_y = y + 0.5;
          float *pixel_color = img + (y * IMAGE_WIDTH + x) * 4;
          float weight = filter_func(outSamples[i].time - 0.5, 0.5);
          weight *= filter_func(outSamples[i].x - center_x, 1);
          weight *= filter_func(outSamples[i].y - center_y, 1);
          weight_sum[x][y] += weight;
          multadd(pixel_color, 1, outSamples[i].color, weight, pixel_color);
        }
      }
    }

    memcpy(img_tmp_out, img, sizeof(float) * 4 * IMAGE_WIDTH * IMAGE_HEIGHT);

    for (int y = 0; y < IMAGE_HEIGHT; y++) {
      for (int x = 0; x < IMAGE_WIDTH; x++) {
        float *pixel_color = img_tmp_out + (y * IMAGE_WIDTH + x) * 4;
        pixel_color[0] = pixel_color[0] / weight_sum[x][y];
        pixel_color[1] = pixel_color[1] / weight_sum[x][y];
        pixel_color[2] = pixel_color[2] / weight_sum[x][y];
        pixel_color[3] = 1;
      }
    }

    /* for debug only
    if (render_y == 256) {
      for (int x = 0; x < IMAGE_WIDTH; x++) {
        float *pixel_color = img_tmp_out + (render_y * IMAGE_WIDTH + x) * 4;
        cout << "color " << pixel_color[0] << endl;
      }
      break;
    }
    */

    //char out_filename[1024];
    //sprintf(out_filename, "br_tmp/brute_force_tmp.exr", render_y);
    WriteExr(out_filename, img_tmp_out, IMAGE_WIDTH, IMAGE_HEIGHT);
  }

  WriteExr(out_filename, img_tmp_out, IMAGE_WIDTH, IMAGE_HEIGHT);

  delete[] img;
  delete[] img_tmp_out;
}

