// Copyright (c) 2009, Yu-Ting Tseng and Kevin Egan
// MIT License, see LICENSE file for details


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <vector>
#include <fstream>
#include <sstream>
#include <iostream>

#include <sys/types.h>
#include <unistd.h>

#include "util.h"
#include "const.h"

using namespace std;

static const int BATCH_SIZE = 500000;





void CallPixieBatch(const vector<InputSample> & inSamples,
    int width, int height, const char* rib_filename,
    const char* infilename, const char* outfilename,
    bool re_run, bool keep_temp_files, int num_batch, int bi);

void ReadPixieBatch(int total_samples,
    vector<FreqSample> & outSamples, int width, int height,
    const char* rib_filename, const char* outfilename,
    bool re_run, bool keep_temp_files, int num_batch, int bi);





// animationOffset says where to start the shutter in the global animation,
// timeCompress scales time between the animation time units and the
// shutter which goes from 0 to SHUTTER_END (sampling happens between
// TIME_START and TIME_END)
void RenderFreqSamplesMultiFile(vector<InputSample> & inSamples,
    vector<FreqSample> & outSamples, int width, int height,
    const char* rib_filename_base,
    const char* infilename_base, const char* outfilename_base, bool re_run,
    bool keep_temp_files, float animationOffset, float timeCompress)
{

  int firstRib = (int) (animationOffset + TIME_START / timeCompress);
  int lastRib = (int) (animationOffset + TIME_END / timeCompress);
  int num_samples = inSamples.size();

  vector<int> frameNumSamples;
  frameNumSamples.resize(lastRib + 1);
  
  cout << "*** RENDER  total: " << inSamples.size() << ", per pix " <<
      ((float) inSamples.size()) / (IMAGE_WIDTH * IMAGE_HEIGHT) << endl;
  for (int frame = firstRib; frame <= lastRib; frame++) {
      vector<InputSample> ribSamples;

      // filter out samples not in the range covered by this rib file
      // frameMin and frameMax are in "shutter" time coordinates
      float frameMinTime = (frame - animationOffset) * timeCompress;
      float frameMaxTime = (frame + 1 - animationOffset) * timeCompress;
      for (int j = 0; j < num_samples; j++) {
        if (inSamples[j].time >= frameMinTime && inSamples[j].time < frameMaxTime) {
            InputSample scaledSample = inSamples[j];
            scaledSample.time = (scaledSample.time - frameMinTime) / (frameMaxTime - frameMinTime);
            ribSamples.push_back(scaledSample);
        }
      }

      frameNumSamples[frame] = ribSamples.size();

      cout << "calling pixie: frame " << frame << ", frame total: " << ribSamples.size() << endl;
      int num_batch = (int) ceil(((float) frameNumSamples[frame]) / BATCH_SIZE);
      for (int bi = 0; bi < num_batch; bi++) {
        // out and in are labeled with respect to pixie
        char infilename[1024];
        char outfilename[1024];
        char rib_filename[1024];
        sprintf(infilename, "%s.%d.%04d.%02d", infilename_base, getpid(), frame, bi);
        sprintf(outfilename, "%s.%d.%04d.%02d", outfilename_base, getpid(), frame, bi);
        sprintf(rib_filename, "%s_%04d.rib", rib_filename_base, frame);

        CallPixieBatch(ribSamples, width, height, rib_filename,
            infilename, outfilename, re_run, keep_temp_files,
            num_batch, bi);
      }
  }

  int total_samples = inSamples.size();
  inSamples.clear();
  outSamples.reserve(total_samples);

  for (int frame = firstRib; frame <= lastRib; frame++) {
      int num_batch = (int) ceil(((float) frameNumSamples[frame]) / BATCH_SIZE);
      for (int bi = 0; bi < num_batch; bi++) {
        // out and in are labeled with respect to pixie
        char outfilename[1024];
        char rib_filename[1024];
        sprintf(outfilename, "%s.%d.%04d.%02d", outfilename_base, getpid(), frame, bi);
        sprintf(rib_filename, "%s_%04d.rib", rib_filename_base, frame);

        int startSamples = outSamples.size();
        ReadPixieBatch(frameNumSamples[frame], outSamples, width, height, rib_filename,
            outfilename, re_run, keep_temp_files, num_batch, bi);

        int endSamples = outSamples.size();

        for (int j = startSamples; j < endSamples; j++) {
            outSamples[j].time = (outSamples[j].time + frame - animationOffset) * timeCompress;
            outSamples[j].v2d_motion = outSamples[j].v2d_motion / timeCompress; 
            outSamples[j].a_min = outSamples[j].a_min / timeCompress; 
            outSamples[j].a_max = outSamples[j].a_max / timeCompress; 
        }
      }
  }
}



void RenderColorSamplesMultiFile(vector<InputSample> & inSamples,
    vector<ColorSample> & outSamples, int width, int height,
    const char* rib_filename_base,
    const char* infilename_base, const char* outfilename_base, bool re_run,
    bool keep_temp_files, float animationOffset, float timeCompress)
{

  int firstRib = (int) (animationOffset + TIME_START / timeCompress);
  int lastRib = (int) (animationOffset + TIME_END / timeCompress);
  int num_samples = inSamples.size();

  vector<int> frameNumSamples;
  frameNumSamples.resize(lastRib + 1);

  //printf("time span %.3f - %.3f, rib span %d - %d, offset %.3f, scale %.3f\n",
  //  TIME_START, TIME_END, firstRib, lastRib, animationOffset, timeCompress);
  cout << "*** RENDER  total: " << inSamples.size() << ", per pix " <<
      ((float) inSamples.size()) / (IMAGE_WIDTH * IMAGE_HEIGHT) << endl;
  
  for (int frame = firstRib; frame <= lastRib; frame++) {
      vector<InputSample> ribSamples;

      // filter out samples not in the range covered by this rib file
      // frameMin and frameMax are in "shutter" time coordinates
      float frameMinTime = (frame - animationOffset) * timeCompress;
      float frameMaxTime = (frame + 1 - animationOffset) * timeCompress;
      for (int j = 0; j < num_samples; j++) {
        if (inSamples[j].time >= frameMinTime && inSamples[j].time < frameMaxTime) {
            InputSample scaledSample = inSamples[j];
            scaledSample.time = (scaledSample.time - frameMinTime) / (frameMaxTime - frameMinTime);
            ribSamples.push_back(scaledSample);
        }
      }

      frameNumSamples[frame] = ribSamples.size();

      int num_batch = (int) ceil(((float) frameNumSamples[frame]) / BATCH_SIZE);
      for (int bi = 0; bi < num_batch; bi++) {
        cout << "calling pixie: frame " << frame << ", frame total: " << ribSamples.size() << endl;
        // out and in are labeled with respect to pixie
        char infilename[1024];
        char outfilename[1024];
        char rib_filename[1024];
        sprintf(infilename, "%s.%d.%04d.%02d", infilename_base, getpid(), frame, bi);
        sprintf(outfilename, "%s.%d.%04d.%02d", outfilename_base, getpid(), frame, bi);
        sprintf(rib_filename, "%s_%04d.rib", rib_filename_base, frame);

        CallPixieBatch(ribSamples, width, height, rib_filename,
            infilename, outfilename, re_run, keep_temp_files,
            num_batch, bi);
      }
  }

  int total_samples = inSamples.size();
  inSamples.clear();
  outSamples.reserve(total_samples);

  vector<FreqSample> tmpFreqVec;

  for (int frame = firstRib; frame <= lastRib; frame++) {
      int num_batch = (int) ceil(((float) frameNumSamples[frame]) / BATCH_SIZE);
      for (int bi = 0; bi < num_batch; bi++) {
        tmpFreqVec.clear();
        tmpFreqVec.reserve(BATCH_SIZE);

        // out and in are labeled with respect to pixie
        char outfilename[1024];
        char rib_filename[1024];
        sprintf(outfilename, "%s.%d.%04d.%02d", outfilename_base, getpid(), frame, bi);
        sprintf(rib_filename, "%s_%04d.rib", rib_filename_base, frame);

        ReadPixieBatch(frameNumSamples[frame], tmpFreqVec, width, height, rib_filename,
            outfilename, re_run, keep_temp_files, num_batch, bi);

        int batch_samples = tmpFreqVec.size();

        // convert from the FreqSample struct to the smaller ColorSample
        for (int i = 0; i < batch_samples; i++) {
            const FreqSample & freqSample = tmpFreqVec[i];
            ColorSample colorSample;

            colorSample.x = freqSample.x;
            colorSample.y = freqSample.y;
            colorSample.time = (freqSample.time + frame - animationOffset) * timeCompress;
            colorSample.density = -1;
            for (int j = 0; j < 4; j++) {
                colorSample.color[j] = freqSample.color[j];
            }
            outSamples.push_back(colorSample);
        }
      }
  }
}










// This returns outSamples that only have color content (less memory),
// used for the second pass
void RenderColorSamples(vector<InputSample> & inSamples,
    vector<ColorSample> & outSamples, int width, int height,
    const char* rib_filename, const char* infilename_base,
    const char* outfilename_base, bool re_run, bool keep_temp_files) {

  int num_batch = (int) ceil(((float) inSamples.size()) / BATCH_SIZE);
  for (int bi = 0; bi < num_batch; bi++) {
    cout << "calling pixie: " << bi * BATCH_SIZE << ", total: " << inSamples.size() << endl;
    // out and in are labeled with respect to pixie
    char infilename[1024];
    char outfilename[1024];
    sprintf(infilename, "%s.%d.%02d", infilename_base, getpid(), bi);
    sprintf(outfilename, "%s.%d.%02d", outfilename_base, getpid(), bi);

    CallPixieBatch(inSamples, width, height, rib_filename,
        infilename, outfilename, re_run, keep_temp_files, num_batch, bi);
  }

  int total_samples = inSamples.size();
  inSamples.clear();
  outSamples.reserve(total_samples);

  vector<FreqSample> tmpFreqVec;

  for (int bi = 0; bi < num_batch; bi++) {
    tmpFreqVec.clear();
    tmpFreqVec.reserve(BATCH_SIZE);

    // out and in are labeled with respect to pixie
    char outfilename[1024];
    sprintf(outfilename, "%s.%d.%02d", outfilename_base, getpid(), bi);

    ReadPixieBatch(total_samples, tmpFreqVec, width, height, rib_filename,
        outfilename, re_run, keep_temp_files, num_batch, bi);

    int num_samples = tmpFreqVec.size();

    // convert from the FreqSample struct to the smaller ColorSample
    for (int i = 0; i < num_samples; i++) {
        const FreqSample & freqSample = tmpFreqVec[i];
        ColorSample colorSample;

        colorSample.x = freqSample.x;
        colorSample.y = freqSample.y;
        colorSample.time = freqSample.time;
        colorSample.density = -1;
        for (int j = 0; j < 4; j++) {
            colorSample.color[j] = freqSample.color[j];
        }
        outSamples.push_back(colorSample);
    }
  }
}










// This returns outSamples that have frequency content, used for the first
// pass
void RenderFreqSamples(vector<InputSample> & inSamples,
    vector<FreqSample> & outSamples, int width, int height,
    const char* rib_filename, const char* infilename_base,
    const char* outfilename_base, bool re_run, bool keep_temp_files) {

  int num_batch = (int) ceil(((float) inSamples.size()) / BATCH_SIZE);
  for (int bi = 0; bi < num_batch; bi++) {
    cout << "calling pixie: " << bi * BATCH_SIZE << ", total: " << inSamples.size() << endl;
    // out and in are labeled with respect to pixie
    char infilename[1024];
    char outfilename[1024];
    sprintf(infilename, "%s.%d.%02d", infilename_base, getpid(), bi);
    sprintf(outfilename, "%s.%d.%02d", outfilename_base, getpid(), bi);

    CallPixieBatch(inSamples, width, height, rib_filename,
        infilename, outfilename, re_run, keep_temp_files, num_batch, bi);
  }

  int total_samples = inSamples.size();
  inSamples.clear();
  outSamples.reserve(total_samples);

  for (int bi = 0; bi < num_batch; bi++) {
    // out and in are labeled with respect to pixie
    char outfilename[1024];
    sprintf(outfilename, "%s.%d.%02d", outfilename_base, getpid(), bi);

    ReadPixieBatch(total_samples, outSamples, width, height, rib_filename,
        outfilename, re_run, keep_temp_files, num_batch, bi);

  }
}




void CallPixieBatch(const vector<InputSample> & inSamples,
    int width, int height, const char* rib_filename,
    const char* infilename, const char* outfilename,
    bool re_run, bool keep_temp_files, int num_batch, int bi)
{

    int num_samples = BATCH_SIZE;
    if (bi == num_batch - 1)
      num_samples = inSamples.size() - bi * BATCH_SIZE;

    // write sample locations
    if (re_run) {
      FILE* infile = fopen(infilename, "w");
      if (infile == NULL) {
        fprintf(stderr, "error opening file for writing '%s'\n", infilename);
        exit(1);
      }

      // write the number of samples
      int writesize = fwrite(& num_samples, sizeof(int), 1, infile);
      if (writesize != 1) {
        fprintf(stderr, "error writing to file '%s', %d\n", infilename, writesize);
        exit(1);
      }

      float* outdata = new float[num_samples * 3];

      // write the data
      for (int i = 0; i < num_samples; i++) {
        outdata[i * 3 + 0] = inSamples[bi * BATCH_SIZE + i].x;
        outdata[i * 3 + 1] = inSamples[bi * BATCH_SIZE + i].y;
        outdata[i * 3 + 2] = inSamples[bi * BATCH_SIZE + i].time;
      }
      writesize = fwrite(outdata, sizeof(float), num_samples * 3, infile);
      if (writesize != num_samples * 3) {
        fprintf(stderr, "error writing to file '%s', %d %d\n",
            infilename, writesize, sizeof(float) * num_samples * 3);
        exit(1);
      }

      delete[] outdata;
      fclose(infile);

      // render with stripped down motion calculation shaders
      char buf[1024];
      sprintf(buf, "~/pixie/install/bin/rndr %s --samplefiles %s %s",
          rib_filename, infilename, outfilename);
      int ret = 0;
      if (re_run)
        ret = system(buf);

      if (ret != 0) {
        fprintf(stderr, "error running pixie. error %d\n", ret);
        exit(1);
      }
      if (!keep_temp_files)
        unlink(infilename);
    }
}



void ReadPixieBatch(int total_samples,
    vector<FreqSample> & outSamples, int width, int height,
    const char* rib_filename, const char* outfilename,
    bool re_run, bool keep_temp_files, int num_batch, int bi)
{
    int num_samples = BATCH_SIZE;
    if (bi == num_batch - 1)
      num_samples = total_samples - bi * BATCH_SIZE;

    // read rendering output
    FILE* outfile = fopen(outfilename, "r");
    if (outfile == NULL) {
      fprintf(stderr, "error opening file for reading '%s'\n", outfilename);
      exit(1);
    }
    int numitems = -1;
    int numread = fread(& numitems, sizeof(int), 1, outfile);

    if (numread != 1) {
      fprintf(stderr, "error reading from file '%s'\n", outfilename);
      exit(1);
    }
    if (numitems != num_samples) {
      fprintf(stderr, "input and output don't match sizes '%s', %d %d\n",
        outfilename, numitems, num_samples);
      exit(1);
    }

    // x, y, time, color, depth, 2 user defined vector outs
    const int floatsPerSample = 3 + 4 + 1 + (2 * 3);
    float* indata = new float[floatsPerSample * num_samples];
    numread = fread(indata, sizeof(float), floatsPerSample * num_samples, outfile);
    fclose(outfile);
    
    if (numread != floatsPerSample * num_samples) {
      fprintf(stderr, "error reading from file '%s', %d %d %d\n",
          outfilename, numread, floatsPerSample, num_samples);
      exit(1);
    }

    for (int s = 0; s < num_samples; s++) {
      FreqSample cursample;
      float bounds[3];

      cursample.x = indata[s * floatsPerSample + 0];
      cursample.y = indata[s * floatsPerSample + 1];
      cursample.time = indata[s * floatsPerSample + 2];
      cursample.color[0] = indata[s * floatsPerSample + 3];
      cursample.color[1] = indata[s * floatsPerSample + 4];
      cursample.color[2] = indata[s * floatsPerSample + 5];
      cursample.color[3] = indata[s * floatsPerSample + 6];
      cursample.zDepth = indata[s * floatsPerSample + 7];
      cursample.motion[0] = indata[s * floatsPerSample + 8];
      cursample.motion[1] = indata[s * floatsPerSample + 9];
      cursample.motion[2] = indata[s * floatsPerSample + 10];
      bounds[0] = indata[s * floatsPerSample + 11];
      bounds[1] = indata[s * floatsPerSample + 12];
      bounds[2] = indata[s * floatsPerSample + 13];

      if (cursample.color[3] == 0.0) {
        // HACK: sample did not hit any surface, fill it in
        // as opaque distant solid black surface
        cursample.color[3] = 1.0;
        cursample.motion[0] = 0.0;
        cursample.motion[1] = 0.0;
        cursample.motion[2] = 0.0;
        bounds[0] = 0.0;
        bounds[1] = 0.0;
        bounds[2] = 0.0;
      }
      cursample.a_min = bounds[0];
      cursample.a_max = bounds[1];
      cursample.omega_x_max = bounds[2];

      float a_mid = 0.5 * (cursample.a_max + cursample.a_min);
      cursample.v2d_motion = Vector2D(cursample.motion[0] * a_mid,
          cursample.motion[1] * a_mid);

      outSamples.push_back(cursample);
    }
    delete[] indata;
    if (!keep_temp_files)
      unlink(outfilename);
}




void HackFreqSamples(vector<FreqSample> & outSamples,
    const char* rib_filename_base, int numSamples)
{
    char motionFile[1024];
    char shadFile[1024];

    for (int i = 0; i < numSamples; i++) {
        int width, height;
        sprintf(motionFile, "%s_%d_motion.exr", rib_filename_base, i + 1);
        sprintf(shadFile, "%s_%d_shad.exr", rib_filename_base, i + 1);
        float* motionData;
        float* shadData;
        ReadExr(motionFile, motionData, width, height);
        ReadExr(shadFile, shadData, width, height);

        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {

              FreqSample cursample;

              cursample.x = x + 0.5;
              cursample.y = y + 0.5;
              cursample.time = (i + 0.5) / numSamples;

              cursample.color[0] = -1;
              cursample.color[1] = -1;
              cursample.color[2] = -1;
              cursample.color[3] = -1;
              cursample.zDepth = -1;

              cursample.motion[0] = motionData[(y * width + x) * 4 + 0];
              cursample.motion[1] = motionData[(y * width + x) * 4 + 1];
              cursample.motion[2] = motionData[(y * width + x) * 4 + 2];

              cursample.a_min = shadData[(y * width + x) * 4 + 0];
              cursample.a_max = shadData[(y * width + x) * 4 + 1];
              cursample.omega_x_max = shadData[(y * width + x) * 4 + 2];

              // TODO move this code into ConstructSegments()
              float a_mid = 0.5 * (cursample.a_max + cursample.a_min);
              cursample.v2d_motion = Vector2D(cursample.motion[0] * a_mid,
                  cursample.motion[1] * a_mid);

              outSamples.push_back(cursample);
            }

        }

        delete[] motionData;
        delete[] shadData;
    }
}




