
#ifndef __MBSHADOW_UTIL__
#define __MBSHADOW_UTIL__

#include <math.h>
#include <stdlib.h>
#include <set>
#include <vector>
#include <utility>

#include "const.h"
#include "config.h"

#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define CLAMP(a, low, high) (MIN(MAX(a, low), high))

struct InputSample {
    float x;
    float y;
    float time;
};


class HaltonSampler
{
public:
    HaltonSampler() {
        _nSample = 0;
        for (int i = 0; i < 3; i++ ) {
            _offset[i] = s_centeroffset[0][i];
        }
    };

    const float* GetOffset() { return _offset; }

    // make the offset such that the 3rd dimension averages to 0
    void CenterSequence(int numSamples) {
        if (numSamples < 1 || numSamples > 16) {
            return;
        }
        for (int i = 0; i < 3; i++ ) {
            _offset[i] = s_centeroffset[numSamples - 1][i];
        }
    }

    void Randomize() {
        for (int i = 0; i < 3; i++ ) {
            _offset[i] = drand48();
        }
    }

    InputSample GetNextSample() {
        InputSample nextSample;

        // WARNING: assumes .x, .y, and .time are all packed together
        float* outdata = & nextSample.x;

        if (_nSample < 128) {
            outdata[0] = s_precompute[_nSample][0];
            outdata[1] = s_precompute[_nSample][1];
            outdata[2] = s_precompute[_nSample][2];
        } else {
            for (int i = 0; i < 3; i++ ) {
                int seed = _nSample;
                outdata[i] = 0.0;
                float base_inv = 1.0 / s_primes[i];

                while (seed != 0) {
                    int digit = seed % s_primes[i];
                    outdata[i] = outdata[i] + digit * base_inv;
                    base_inv = base_inv / s_primes[i];
                    seed = seed / s_primes[i];
                }
            }
        }

        for (int i = 0; i < 3; i++ ) {
            outdata[i] += _offset[i];
            outdata[i] -= (outdata[i] >= 1.0) ? 1.0 : 0.0;
        }

        _nSample++;

        return nextSample;
    }

private:
    int _nSample;
    float _offset[3];

    static const int s_primes[3];

    static const float s_centeroffset[16][3];
    static const float s_precompute[128][3];
};



// Define a 2d vector class support various operators like +, -, *, inner
// product...
class Vector2D {
 public:
  float x;
  float y;

  static const float EPSILON = 0.00001;

  Vector2D() {
    x = 0;
    y = 0;
  }

  Vector2D(float x, float y) {
    this->x = x;
    this->y = y;
  }

  Vector2D(float *array) {
    this->x = array[0];
    this->y = array[1];
  }

  float length() const {
    float l2 = lengthSquared();
    return l2 == 0 ? 0 : sqrt(l2);
  }

  float lengthSquared() const {
    return x * x + y * y;
  }

  float angle() const {
    if (x == 0 && y == 0)
      return 0;
    else if (y > 0)
      return acos(x / length());
    else
      return 2 * M_PI - acos(x / length());
  }

  bool operator== (const Vector2D& v) {
    return fabs(x - v.x) < EPSILON &&
           fabs(y - v.y) < EPSILON;
  }


  Vector2D *operator= (const float *array) {
    x = array[0];
    y = array[1];
    return this;
  }

  Vector2D operator+ (const Vector2D& v) {
    return Vector2D(x + v.x, y + v.y);
  }

  Vector2D operator- (const Vector2D& v) {
    return Vector2D(x - v.x, y - v.y);
  }

  Vector2D operator* (float s) {
    return Vector2D(x * s, y * s);
  }

  Vector2D operator/ (float s) {
    return Vector2D(x / s, y / s);
  }

  Vector2D UnitVector() const {
    float length = this->length();
    if (length == 0)
      return Vector2D(0, 0);
    else
      return Vector2D(x / length, y / length);
  }

  float DotProduct(const Vector2D& v) const {
    return x * v.x + y * v.y;
  }

  static Vector2D Zero() { return Vector2D(0, 0); }
};

inline std::ostream &operator<< (std::ostream& out,
                                 Vector2D& v) {
  out << "(" << v.x << ", " << v.y << ") ";
  return out;
}


// we use class inheritance to reuse code, FreqSample stores a super-set
// of information relative to ColorSample
class ColorSample {
public:
    float x;
    float y;
    float time;
    float density;

    float color[4];
};


class FreqSample : public ColorSample {
public:
    float zDepth;
    float shift_time;
    float motion[3];
    Vector2D v2d_motion;
    float omega_x_max;
    float a_min;
    float a_max;
};




struct TimeSegmentAtPixel {
  float start_time;
  float end_time;
  float time_interval;
  float x;
  float y;
  std::vector<FreqSample*> samples;
  Vector2D avg_motion;
  float omega_x_max;
  float a_min;
  float a_max;

  float wide_filter_radius;
  float refined_filter_radius;

  // we do a greedy algorithm to see if we can use the widest
  // filter radius possible
  bool filter_calculated;  // this filter has been manually checked
  bool filter_marked;      // manually checked or marked by the greedy algorithm
};

struct Filter {
  Vector2D avg_motion;
  Vector2D max_motion;
  Vector2D min_motion;
};

void ReadExr(const char* inputFile, float* & inData, int & width, int & height);
void WriteExr(const char* outputFile, const float* outData, int width, int height);
void ReadRaw(const char* inputFile, float* & inData, int & width, int & height);
void WriteRaw(const char* outputFile, const float* outData, int width, int height);
void ReadTxt(const char* inputFile, const float* inData, int width, int height);
void WriteTxt(const char* outputFile, const float* outData, int width, int height);
void WriteRgb(const char* outputFile, const float* outData, int width, int height);

// h (hue) should be between 0.0 and 1.0
void RainbowtoRGB( float *r, float *g, float *b, float h);

// Generate sample locations. Support both uniform and stratefied sampling.
void GenerateSamples(int width, int height, float time_start, float time_end,
                     int x_resolution, int y_resolution, float time_resolution,
                     int left, int right, int top, int bottom,
                     bool stratified, std::vector<InputSample> &samples,
                     bool vertical = false);

// Given a set of samples, find the max/min/representative motion.
Vector2D FindAvgMotion(std::vector<FreqSample*>& samples);
float FindMaxMotionLength(std::vector<FreqSample*>& samples);
float FindMinMotionLength(std::vector<FreqSample*>& samples);
float FindMaxOmegaXMax(std::vector<FreqSample*>& samples);

// Frequency content measured in pixels and time units.
// In this implementation we sample at least once per pixel per shutter
// so we can ignore very slow moving very low frequency signals.
inline bool HasSomeFrequencyContent(float omega_x_max, float a_max)
{

#if 0
    // FIXME
    // not implemented before submission
    //
    // when we read in omega_x_max values we clamp to a minimum of 0.2
    // to prevent filters from extending to far based on local freq info
    return (omega_x_max > 0.25 || a_max * omega_x_max > 0.25);
#else
    return true;
#endif
}

inline bool IsConsistentDirections(const Vector2D& a, const Vector2D& b) {
  //const float angle_tolerance = 20.0 * M_PI / 180;
  //return fabs(fmod(a.angle() - b.angle() + 2 * M_PI + M_PI, 2 * M_PI) - M_PI) < angle_tolerance;

  // 0.940 - cos(20 * PI / 180)
  const float dot_cutoff = 0.940;;
  return (a.DotProduct(b) / (a.length() * b.length())) > dot_cutoff;
}


bool IsFilterOutOfImageBound(TimeSegmentAtPixel *segment);

bool FilterSamplesConsistent(TimeSegmentAtPixel *segment,
                        std::vector<FreqSample*>* samples_in_grid,
                        const std::vector<FreqSample*> & candidate_samples);

// Find neighbor samples that can be used to estimate a_min and a_max for a
// pixel. The input is a set of samples indexed by x and y coordinates.
void CollectSamples(std::vector<FreqSample*>* samples_in_grid,
                    Vector2D center,
                    std::vector<FreqSample*>* out_samples);

// Define the function pointer of a filter function.
typedef float (*FilterFunc)(float, float);

// Check if a sample is in the reconstruction filter. If yes, write the
// function value to *weight.
bool IsInFilter(TimeSegmentAtPixel *segment, ColorSample& sample,
                FilterFunc, float *weight);

// Decide whether two velocity vectors are similar.
bool compareVelocities(const float* vel1, const float* vel2);

// Check whether there is aliasing.
bool IsAliasing(float time_sampling_rate, float section_time_interval,
                float a_min, float a_max);

// Find nearby samples that are possibly covered by a filter.
void GetFilterFreqSamples(const TimeSegmentAtPixel& segment,
                      std::vector<FreqSample*>* samples_in_grid,
                      std::vector<FreqSample*>* output);
void GetFilterColorSamples(const TimeSegmentAtPixel& segment,
                      std::vector<ColorSample*>* samples_in_grid,
                      std::vector<ColorSample*>* output);

void GetNearbyPixels(const TimeSegmentAtPixel& segment,
                     std::vector<std::pair<int, int> >* nearby_pixels,
                     const Vector2D& unit_motion,
                     float motionStart, float motionEnd,
                     float perpStart, float perpEnd);
void GetFilterPixels(const TimeSegmentAtPixel& segment,
                     std::vector<std::pair<int, int> >* nearby_pixels);

void FillSamplingRateToNearbyPixels(const TimeSegmentAtPixel& segment,
    float temporal_sampling_rate, float spatial_sampling_rate,
    float *temporal_sampling_rate_dest, float *spatial_sampling_rate_dest);

// Find samples within a filter and calculate their weights. The outputs are
// <sample, weight> pairs.
void ApplyFilter(TimeSegmentAtPixel& segment,
                 FilterFunc filter_func,
                 std::vector<ColorSample*>* samples_in_grid,
                 std::vector<std::pair<ColorSample*, float> >* output,
                 bool useBucketNormalization,
                 bool oldBucketNormalization);

// Assume each pixel is 4-byte long
void ClearImage(float *img, int width, int height);


// Blur amin and amax
void BlurMotion(std::vector<TimeSegmentAtPixel*>* segments, int radius);

// Define various filter functions.
inline float SincWindow(float x, float width) {
  if (fabs(x) < 0.000001)
    return 1;
  else
    return sin(M_PI * x / width) / (M_PI * x / width);
}



// we just use the windowed gaussian
//inline float Gaussian(float x, float std) {
//  return exp(- x * x / (2 * std * std));
//}


// Use a Gaussian function.  Currently we cutoff after one "lobe",
// and one lobe for a gaussian is 2 standard deviations (read below).
// Furthermore we've set spatial filters to have 1 lobe be one pixel
// long, so without motion, our spatial filter should have a 1 pixel
// radius, and it should be a gaussian that ends at two standard
// deviations.
//
// the width parameter is the same as the refined_filter_radius radius
// and other measures (expressed in units 1 / (2 * bandlimit)),
// for a spatial reconstruction sinc filter the radius is the pixel
// width of a lobe, it is also half a wavelength that the filter can capture.
//
// this function needs to have a scale k such that
//   width * k = std deviation length
//
// NOT CURRENTLY USED
//   fourier transform of gaussian has stdlen = (1 / (2pi * dualSpaceDev))
//   if we say in freq space that a stdlen = 1 gaussian has a bandlimit
//     of unit 1, then that gaussian in the primal domain has
//     stdlen (1 / (2pi))
// 
//   radius (width parameter) = 1 / (2 * bandlimit)
//     = 1 / 2
//   stdlen = 1 / (2pi)
//   stdlen = radius / pi
//
// CURRENTLY USED
//   Similarly if we have a pair (stdlen, bl) that we like, then we get
//   new bl' the question is what is stdlen'?
//
//   r' (radius' aka the new width parameter) = 1 / (2 * bl')
//   scaling = bl' / bl = (1 / (2r')) / bl
//   stdlen' = stdlen * (1 / scaling)
//     = stdlen * (bl / (1 / 2r')) = 2r'(bl)(stdlen)
//   new stddev = (width parameter) * (verified bandlimit) * (verified stddev)
//
// the LOBES parameter would control how many radii you want, to convert
// LOBES to stddev you need to do
//   totalwidth = LOBES * radius = LOBES * (stddev * scale)
//   totalwidth = stddev * (LOBES / scale)
inline float GaussianWindow(float x, float width) {
  // TODO the ratio between two standard deviations to a lobe is baked
  // into the gaussian_lookup table
  //
  // float std = width / 2.0;

    static const float gaussian_lookup[65] = { 0.86466, 0.86418, 0.86271, 0.86028, 0.85688, 0.85253, 0.84724, 0.84102, 0.83390, 0.82589, 0.81701, 0.80729, 0.79677, 0.78546, 0.77340, 0.76062, 0.74716, 0.73306, 0.71834, 0.70306, 0.68724, 0.67094, 0.65419, 0.63703, 0.61950, 0.60166, 0.58353, 0.56517, 0.54661, 0.52789, 0.50905, 0.49014, 0.47120, 0.45225, 0.43334, 0.41450, 0.39576, 0.37716, 0.35873, 0.34050, 0.32250, 0.30474, 0.28727, 0.27008, 0.25322, 0.23670, 0.22053, 0.20473, 0.18932, 0.17430, 0.15969, 0.14549, 0.13172, 0.11837, 0.10546, 0.09297, 0.08093, 0.06932, 0.05815, 0.04740, 0.03709, 0.02719, 0.01772, 0.00866, 0.00000 };

    // sample is between 0 and 1 lobe, which is 0 to 2 std dev.
    float sample = fabs(x) / width;
    if (sample > 0.9999) {
        return 0.0; 
    }
    float scaled = sample * 64;
    int index = (int) scaled;
    float weight = scaled - index;
    return (1.0 - weight) * gaussian_lookup[index] + weight * gaussian_lookup[index + 1];
}



// returns 0 if any of the 3 filters are zero
inline bool In3DGaussianWindow(float x, float xwidth, float y, float ywidth, float t, float twidth) {
    return (fabs(t) < twidth && fabs(x) < xwidth && fabs(y) < ywidth);
}

// Calculate how many temporal samples are needed to avoid aliasing.
float NumTemporalSamplesNeeded(TimeSegmentAtPixel* segment);

// Calculate spatial sampling rate needed to avoid aliasing.
float NumSpatialSamplesNeeded(TimeSegmentAtPixel* segment);



float GetOptimalFilterWidth(float a_min, float a_max, float omega_x_max);
float GetFilterBound(float filter_radius);
void GreedyRefineFilterWidth(std::vector<TimeSegmentAtPixel*>* segments,
        int x, int y, std::vector<FreqSample*>* samples_in_grid,
        float maxWidth);
void RefineFilterWidth(TimeSegmentAtPixel *segment,
                        std::vector<FreqSample*>* samples_in_grid,
                        float maxWidth);

// Inline Functions.

inline bool ShouldUseAxisAlignedFilter(const TimeSegmentAtPixel& segment) {
  return segment.a_min < OMEGA_T_MAX / fmin(segment.omega_x_max, OMEGA_PIX);
}

inline float GetOffset(bool stratified) { return stratified ? drand48() : 0.5; }


// Get the filter sheared slope. The unit is pixels / time. See
// doc/figure/filter.pdf for the derivation.
//
// This is inverse of the slope used to compute the shear, which is
// why we divide by a_mid often to compute coordinates inside the sheared
// filter
inline float GetFilterShearedSlope(float a_min, float a_max) {
  if (a_max < 0.00001)
    return 0;
  else
    return a_max * a_min / ((a_max + a_min) / 2);
}


inline void SetZero4(float *a) {
  a[0] = 0;
  a[1] = 0;
  a[2] = 0;
  a[3] = 0;
}

inline void copy(const float* in1, float* out)
{
    out[0] = in1[0];
    out[1] = in1[1];
    out[2] = in1[2];
}


inline void add(const float* in1, const float* in2, float* out)
{
    out[0] = in1[0] + in2[0];
    out[1] = in1[1] + in2[1];
    out[2] = in1[2] + in2[2];
}


inline void subtract(const float* in1, const float* in2, float* out)
{
    out[0] = in1[0] - in2[0];
    out[1] = in1[1] - in2[1];
    out[2] = in1[2] - in2[2];
}


inline void multadd(const float* in1, float scale1, const float* in2, float scale2, float* out)
{
    out[0] = in1[0] * scale1 + in2[0] * scale2;
    out[1] = in1[1] * scale1 + in2[1] * scale2;
    out[2] = in1[2] * scale1 + in2[2] * scale2;
}


inline void negate(float* in1)
{
    in1[0] = -in1[0];
    in1[1] = -in1[1];
    in1[2] = -in1[2];
}


inline void scale(float* in1, float scale)
{
    in1[0] *= scale;
    in1[1] *= scale;
    in1[2] *= scale;
}


inline void cross(const float* in1, const float* in2, float* out)
{
    out[0] =   in1[1] * in2[2] - in1[2] * in2[1];
    out[1] = - in1[0] * in2[2] + in1[2] * in2[0];
    out[2] =   in1[0] * in2[1] - in1[1] * in2[0];
}


inline float dot(const float* in1, const float* in2)
{
    return in1[0] * in2[0] + in1[1] * in2[1] + in1[2] * in2[2];
}


inline float lengthSquared(const float* in1)
{
    return dot(in1, in1);
}


inline float length(const float* in1)
{
  float l2 = lengthSquared(in1);
    return l2 == 0 ? 0 : sqrt(l2);
}


inline void normalize(float* in1)
{
    return scale(in1, 1.0 / length(in1));
}






//////////////////////////////////////////
// 2 element vector functions
//////////////////////////////////////////



inline void copy2(const float* in1, float* out)
{
    out[0] = in1[0];
    out[1] = in1[1];
}


inline void add2(const float* in1, const float* in2, float* out)
{
    out[0] = in1[0] + in2[0];
    out[1] = in1[1] + in2[1];
}


inline void multadd2(const float* in1, float scale1, const float* in2, float scale2, float* out)
{
    out[0] = in1[0] * scale1 + in2[0] * scale2;
    out[1] = in1[1] * scale1 + in2[1] * scale2;
}


inline void subtract2(const float* in1, const float* in2, float* out)
{
    out[0] = in1[0] - in2[0];
    out[1] = in1[1] - in2[1];
}


inline void scale2(float* in1, float scale)
{
    in1[0] *= scale;
    in1[1] *= scale;
}


inline float dot2(const float* in1, const float* in2)
{
    return in1[0] * in2[0] + in1[1] * in2[1];
}


inline float lengthSquared2(const float* in1)
{
    return dot2(in1, in1);
}


inline float length2(const float* in1)
{
  float length_square = lengthSquared2(in1);
  if (length_square != 0)
    return sqrt(length_square);
  else
    return 0;
}


inline void normalize2(float* in1)
{
    return scale2(in1, 1.0 / length2(in1));
}






#endif // __MBSHADOW_UTIL__


